/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Rojo aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * "$Id: out_vorbis.cpp,v 1.10 2002/12/30 08:47:27 jaromil Exp $"
 *
 */

#include <string.h>
#include <inttypes.h>
#include <out_vorbis.h>
#ifdef HAVE_VORBIS

#include <jutils.h>
#include <generic.h>
#include <config.h>

OutVorbis::OutVorbis()
  : OutChannel("vorbis") {
  func("OutVorbis::OutVorbis() %p",this);

  headersize = 0;

  sprintf(name,"Ogg/Vorbis encoder");
  sprintf(version,"version unknown");

  tipo = OGG;

}


bool OutVorbis::init() {
  func("OutVorbis::init() %p",this);
  act("initializing %s %s",name,version);

  vorbis_info_init(&vi);  
  initialized = true;
  if(!apply_profile()) {
    error("problems in setting up codec parameters");
    //    vorbis_info_clear(&vi);
    return false;
  }

  func("vorbis defaults chan %i, freq %u, qual %f)",
       channels(), freq(), quality()/10.0f);
  
  vorbis_encode_setup_init(&vi);

  /* Now, set up the analysis engine, stream encoder, and other
     preparation before the encoding begins. */
  vorbis_analysis_init(&vd,&vi);
  func("vorbis_analysis_init OK");
  vorbis_block_init(&vd,&vb);
  func("vorbis_block_init OK");
  ogg_stream_init(&os, time(NULL));
  func("ogg_stream_init OK");

  /* sets up comments */
  {
    char tmp[128];
    sprintf(tmp,"%s version %s",PACKAGE,VERSION);
    vorbis_comment_init(&vc);
    vorbis_comment_add_tag(&vc,"Streamed with",tmp);
  }

  /* Now, build the three header packets and send through to the stream 
     output stage (but defer actual file output until the main encode loop) */

  /* Build the packets */
  vorbis_analysis_headerout
    (&vd,&vc,&header_main,&header_comments,&header_codebooks);

  /* And stream them out */
  ogg_stream_packetin(&os,&header_main);
  ogg_stream_packetin(&os,&header_comments);
  ogg_stream_packetin(&os,&header_codebooks);
  
  /* write out headers */
  headersize = 0;
  _pbyte = (int8_t*)header; 
  while(ogg_stream_flush(&os,&og)) {
    memcpy(_pbyte,og.header,og.header_len);
    _pbyte += og.header_len;
    memcpy(_pbyte,og.body,og.body_len);
    _pbyte += og.body_len;
    headersize += og.header_len + og.body_len;
  }

  vorbis_comment_clear(&vc);

  initialized = true;
  return initialized;
}

/* note: num is number of samples in the L (or R) channel
   not the total number of samples in pcm[] */
int OutVorbis::encode(short int *pcm, int num) {    
  _pbyte = (int8_t*)buffer;

  int l;

  if(headersize) {
    memcpy(buffer,header,headersize);
    encoded = headersize;
    shout(); dump();
  }
  
  encoded = headersize = 0;

  _intbuf = vorbis_analysis_buffer(&vd,num);

  l = prepare(pcm,_intbuf,num);

  vorbis_analysis_wrote(&vd,l);

  /* While we can get enough data from the library to analyse, one
     block at a time... */
  while(vorbis_analysis_blockout(&vd,&vb)==1) {
    /* Do the main analysis, creating a packet */
    vorbis_analysis(&vb, NULL); /* here as a second argument darkice uses 
				   &op while oggenc uses NULL */
    vorbis_bitrate_addblock(&vb);
    
    while(vorbis_bitrate_flushpacket(&vd, &op)) {
      /* Add packet to bitstream */
      ogg_stream_packetin(&os,&op);
      
      while(ogg_stream_pageout(&os,&og)) { 

	memcpy(_pbyte,og.header,og.header_len);
	_pbyte += og.header_len;
	memcpy(_pbyte,og.body,og.body_len);
	_pbyte += og.body_len;
	
	encoded += og.header_len + og.body_len;
	
	if(ogg_page_eos(&og)) { func("Ogg encoder EOS!"); return encoded; }
      }

    }
  }
  return encoded;
}

int OutVorbis::prepare(short *buf, float **fbuf, int num) {
  int i=0;
  switch(channels()) {
  case 1:
    for(i=0; i<num; i++)
      fbuf[0][i] = (buf[i*2] + buf[i*2+1]) / 65536.0f; // /2 /32768.0f;
    break;
  case 2:
    for(i=0; i<num; i++) {
      fbuf[0][i] = buf[i*2] /32768.0f;
      fbuf[1][i] = buf[i*2+1] /32768.0f;
    }
    break;
  default:
    error("no more than 2 audio channels are supported");
    break;
  }

  return i;

  /*
    the way oggenc does this is by using a 16bit large pointer and shifting around: 
    fbuf[0][i] = ( ( p[i*4+1] << 8 ) | ( p[i*4] & 0xff ) ) / 32768.0f;
    fbuf[1][i] = ( ( p[i*4+3] << 8 ) | ( p[i*4+4] & 0xff ) ) / 32768.0f;
    (or something like this, check it out in oggenc/audio.c to be sure)
  */
}

void OutVorbis::flush() {
  lock();
  vorbis_analysis_wrote(&vd,0); /* oggenc L255 */
  unlock();
}

bool OutVorbis::apply_profile() {
  bool res = true;

  /* TODO
     resampling for ogg vorbis
     meanwhile we force 44100 freq */
  freq(44100);

  func("apply vorbis_encode_setup_vbr(%p,%u,%u,%f)",
       &vi, channels(), freq(), quality()/10.0f);
  if( vorbis_encode_setup_vbr
      (&vi, channels(), freq(), quality()/10.0f) ) {
    error("vorbis_encode_setup_vbr failed: invalid parameters (for quality?)");
    res = false;
  }
  
  /* Turn off management entirely (if it was turned on)
     vorbis_encode_ctl(&vi, OV_ECTL_RATEMANAGE_SET, NULL); */
  
  // } else {
  //  if( vorbis_encode_setup_managed
  //  (&vi, channels(), freq(), -1, bps(), -1 ) ) {
  //	(&vi, 2, SAMPLE_RATE, -1, bps()*1000, -1 ) ) {
  //error("vorbis_encode_setup_managed failed: invalid bitrate %i",bps()*1000);
  // res = false;
  // }
  //  }

  return res;
}
  
OutVorbis::~OutVorbis() {
  func("OutVorbis::~OutVorbis() %p",this);
  if(running) flush();
  act("closing ogg/vorbis encoder");

  ogg_stream_clear(&os);
  
  vorbis_block_clear(&vb);
  vorbis_dsp_clear(&vd);
  vorbis_info_clear(&vi);
}

#endif
