/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Rojo aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * "$Id: outchannels.cpp,v 1.43 2003/03/10 18:14:55 jaromil Exp $"
 *
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <generic.h>
#include <outchannels.h>
#include <jutils.h>
#include <config.h>

#ifdef HAVE_LAME
#include <out_lame.h>
#endif

 #ifdef HAVE_VORBIS
#include <out_vorbis.h>
#endif


/* I think a tighter bound could be:  (mt, March 2000)
 * MPEG1:
 *    num_samples*(bitrate/8)/samplerate + 4*1152*(bitrate/8)/samplerate + 512
 * MPEG2:
 *    num_samples*(bitrate/8)/samplerate + 4*576*(bitrate/8)/samplerate + 256
 */

/* mp3 encoder */


OutChannel::OutChannel(char *myname)
  : Profile(myname) {
  quit = false;
  initialized = false;
  running = false;
  streaming = true;
  sprintf(name,"%s",myname);
  sprintf(version, "  ");
  encoded = 0;
  fd = NULL;
  _thread_initialized = false;
  _thread_init();

  iceid = 0;

  erbapipa = new Pipe(OUT_PIPESIZE);

  profile_changed = true;

  struct settings conf[] = {
    { "bitrate", (void*)&_bps, cfgINT, "0" },
    { "frequency", (void*)&_freq, cfgINT, "0" },
    { "channels", (void*)&_channels, cfgINT, "1" },
    { "quality", (void*)&_quality, cfgFLOAT, "1.0" },
    { "lowpass", (void*)&_lowpass, cfgINT, "0" },
    { "highpass", (void*)&_highpass, cfgINT, "0" },
    { NULL, NULL, cfgNULL, NULL }
  };
  setup(conf);

  if( !load_profile("default") )
    create_default_profile();

}

OutChannel::~OutChannel() {
  func("OutChannel::~OutChannel");

  quit = true; signal();  

  initialized = false;

  jsleep(0,50);

  Shouter *ice = (Shouter*)icelist.begin();
  lock_ice();
  while(ice) {
    icelist.rem(1);
    delete ice;
    ice = (Shouter*)icelist.begin();
  }
  unlock_ice();

  delete erbapipa;

  if(fd) dump_stop();

  /* QUAAAA */
  //_thread_destroy();
}

void OutChannel::start() {
  Shouter *ice = (Shouter*) icelist.begin();
  lock_ice();
  while(ice) {
    if( ice->apply_profile() )
      ice->start();
    ice = (Shouter*) ice->next;
  }
  unlock_ice();
  pthread_create(&_thread, &_attr, &kickoff, this);
}

void OutChannel::run() {
  if(!initialized) {
    warning("OutChannel::run() : output channel uninitialized, thread won't start");
    return;
  }

  running = true;
  while(!quit) {

    encoding = (fd||streaming);

    if(!encoding) {
      jsleep(0,50);
      continue;
    }

    if( erbapipa->read
	(OUT_CHUNK<<2,pcm) <0) {
      jsleep(0,10);
      continue;
    }

    lock();

    encoded = encode(pcm,OUT_CHUNK);

    if(!encoded) { unlock(); continue; }

    calc_bitrate(encoded);

    unlock();

    /* stream it to the net */
    shout();

    /* save it on the harddisk */
    dump();

    /* TODO: flush when erbapipa->read != OUT_CHUNK */
    
  }

  running = false;
}

int OutChannel::create_ice() {
  Shouter *ice = new Shouter();

  
  /* the icecast id is the position in the linklist */
  lock_ice();
  icelist.add(ice);
  ice->id = id + (++iceid); 

  switch(tipo) {

  case MP3:
    ice->format = SHOUT_FORMAT_MP3;
    ice->login(SHOUT_PROTOCOL_XAUDIOCAST);
    break;

  case OGG:
    ice->format = SHOUT_FORMAT_VORBIS;
    ice->login(SHOUT_PROTOCOL_ICE);
    break;

  default:
    error("codec is not streamable");
    delete ice;
    return -1;

  }

  ice->apply_profile();

  unlock_ice();

  return ice->id;
}

void OutChannel::delete_ice(int id) {

  Shouter *ice = get_ice(id);

  if(ice) {
    lock_ice();
    if(ice->running) ice->stop();
    icelist.rem(id);
    delete ice;
    unlock_ice();
  }
  if(!icelist.len() && !fd) encoding = false;
  
}

Shouter *OutChannel::get_ice(int id) {
  return (Shouter*)icelist.pick_id(id);
}

bool OutChannel::apply_ice(int id) {
  bool res = false;
  
  Shouter *ice = get_ice(id);

  if(ice) {
    lock_ice();
    res = ice->apply_profile();
    unlock_ice();
  }
  return res;
}

bool OutChannel::connect_ice(int id, bool on) {
  bool res = false;
  Shouter *ice = get_ice(id);
  if(ice) {
    lock_ice();
    res = (on) ? ice->start() : ice->stop();
    unlock_ice();
  } else 
    error("Outchannel::connect_ice : can't find shouter with id %d");

  return res;
}

int OutChannel::shout() {
  int res, sentout = 0;
  time_t now = time(NULL);
  lock_ice();
  streaming = false;
  Shouter *ice = (Shouter*)icelist.begin();
  while(ice) {
    if(ice->running) {
      streaming = true;
      res = ice->send(buffer,encoded);
      if(res<0) { /* there is an error: -1=temporary , -2=fatal */
	if(res==-1) break;
	if(res==-2) {
	  int errid = icelist.pos(ice);
	  error("fatal error on stream id[%i] to %s:%u",errid,ice->host(),ice->port());
	  ice->stop();
	  ice->retry = now;
	}
      } else sentout += res;
    } else if(ice->retry>0) {
      if((now - ice->retry) > RETRY_DELAY) {
	notice("retry to connect to %s after %u seconds",ice->host(),RETRY_DELAY);
	if( ice->start() ) ice->retry = 0;
	else ice->retry = now;
      }
    }
    ice = (Shouter*)ice->next;
  }
  unlock_ice();
  return sentout;
}

bool OutChannel::dump_start(char *file) {
  struct stat st;
  char temp[MAX_PATH_SIZE];
  int num = 0;
  
  if(fd) {
    warning("%s channel allready dumping to %s",name,fd_name);
    return false;
  }

  /* avoid to overwrite existent files */
  snprintf(temp,MAX_PATH_SIZE,"%s",file);
  while(stat(temp, &st) != -1) {
    /* file EXIST */
    num++;
    snprintf(temp,MAX_PATH_SIZE,"%s.%i",file,num);
  }

  fd = fopen(temp,"wb"); /* writeonly nonblocking binary (-rw-rw-r--) */
  if(!fd) {
    error("%s channel can't open %s for writing",name,temp);
    act("%s",strerror(errno));
    return(false);
  }

  strncpy(fd_name,temp,MAX_PATH_SIZE);
  notice("%s channel dumping to file %s",name,fd_name);

  return true;
}

bool OutChannel::dump_stop() {
  if(!fd) {
    warning("%s channel is not dumping to any file",name);
    return false;
  }

  fflush(fd);
  
  act("%s channel stops dumping to %s",name,fd_name);
  
  fclose(fd);
  fd = NULL;

  return true;
}

bool OutChannel::dump() {
  int res;
  if(!fd) return false;
  fflush(fd);
  res = fwrite(buffer,1,encoded,fd);
  if(res != encoded)
    warning("skipped %u bytes dumping to file %s",encoded - res,fd_name);
  return true;
}

void OutChannel::bps(int in) {
  _bps = in;
  Shouter *ice = (Shouter*)icelist.begin();
  while(ice) {
    ice->_bps = in;
    ice = (Shouter*)ice->next;
  }
}

bool OutChannel::calc_bitrate(int enc) {
  /* calcolates bitrate */
  bytes_accu += enc;
  now = dtime();
  if((now-prev)>1) { /* if one second passed */
    bitrate = (bytes_accu<<3);
    bytes_accu = 0;
    prev = now;
    return true;
  }
  return false;
}

char *OutChannel::guess_bps() {
  int q = (int)fabs(quality());

  switch(q) {

#define BPSVAL(b,f) \
if(!bps()) bps(b); if(!freq()) freq(f); \
snprintf(quality_desc,256,"%uKbit/s %iKhz",b,f/1000);
  
  case 0: BPSVAL(16,11025); break;
  case 1: BPSVAL(16,16000); break;
  case 2: BPSVAL(24,16000); break;
  case 3: BPSVAL(24,22050); break;
  case 4: BPSVAL(32,22050); break;
  case 5: BPSVAL(48,22050); break;
  case 6: BPSVAL(56,22050); break;
  case 7: BPSVAL(64,22050); break;
  case 8: BPSVAL(96,44100); break;
  case 9: BPSVAL(128,44100); break;

  }
  return quality_desc;
}

void OutChannel::push(void *data, int len) {
  
  /* check out if encoders are configured */
  if(!encoding) return;
  
  /* push in data
     wait if pipe is full or occupied
     returns the right thing */
  while( running && 
	 erbapipa->write(len,data) < 0 ) {
    jsleep(0,50);
  }
}

/* thread stuff */

void OutChannel::_thread_init() {
  if(_thread_initialized) return;

  func("OutChannel::thread_init()");
  if(pthread_mutex_init (&_mutex,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_mutex_init (&_mutex_ice,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_cond_init (&_cond, NULL) == -1)
    error("error initializing POSIX thread condition"); 
  if(pthread_attr_init (&_attr) == -1)
    error("error initializing POSIX thread attribute");
  
  /* set the thread as detached
     see: man pthread_attr_init(3) */
  pthread_attr_setdetachstate(&_attr,PTHREAD_CREATE_DETACHED);

  _thread_initialized = true;
}

void OutChannel::_thread_destroy() {
  if(!_thread_initialized) return;
  
  /* we signal and then we check the thread
     exited by locking the conditional */
  if(running) {
    signal();
    lock(); unlock();
  }

  if(pthread_mutex_destroy(&_mutex) == -1)
    error("error destroying POSIX thread mutex");
  if(pthread_cond_destroy(&_cond) == -1)
    error("error destroying POSIX thread condition");
  if(pthread_attr_destroy(&_attr) == -1)
    error("error destroying POSIX thread attribute");
  _thread_initialized = false;
}
